@@ -13,19 +13,32 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ *
+ *  In addition, as a special exception, the copyright holders give permission
+ *  to link the code of portions of this program with the OpenSSL library under
+ *  certain conditions as described in each individual source file, and
+ *  distribute linked combinations including the two.
+ *  You must obey the GNU General Public License in all respects for all of the
+ *  code used other than OpenSSL.  If you modify file(s) with this exception,
+ *  you may extend this exception to your version of the file(s), but you are
+ *  not obligated to do so.  If you do not wish to do so, delete this exception
+ *  statement from your version.  If you delete this exception statement from
+ *  all source files in the program, then also delete it here.
  */
 
 #include "config.h"
 #include "log.h"
 
+#include <openssl/x509.h>  /* work around OpenSSL bug: missing definition of STACK_OF */
+#include <openssl/tls1.h>
+
+#include <sys/stat.h>
+
+#include <ctype.h>
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <ctype.h>
-#include <sys/stat.h>
-#include <limits.h>
-#include <openssl/x509.h>  /* work around OpenSSL bug: missing definition of STACK_OF */
-#include <openssl/tls1.h>
 
 const struct vpn_config invalid_cfg = {
 	.gateway_host = {'\0'},
@@ -40,6 +53,9 @@ const struct vpn_config invalid_cfg = {
 	.set_routes = -1,
 	.set_dns = -1,
 	.pppd_use_peerdns = -1,
+#if HAVE_RESOLVCONF
+	.use_resolvconf = -1,
+#endif
 	.use_syslog = -1,
 	.half_internet_routes = -1,
 	.persistent = -1,
@@ -109,6 +125,7 @@ int strtob(const char *str)
 		return -1;
 
 	long i = strtol(str, NULL, 0);
+
 	if (i < 0 || i > 1)
 		return -1;
 	return i;
@@ -125,21 +142,21 @@ int parse_min_tls(const char *str)
 	if (str[0] != '1' || str[1] != '.' || str[2] == 0 || str[3] != 0)
 		return -1;
 	switch (str[2]) {
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#ifdef TLS1_VERSION
 	case '0':
 		return TLS1_VERSION;
+#endif
+#ifdef TLS1_1_VERSION
 	case '1':
 		return TLS1_1_VERSION;
+#endif
+#ifdef TLS1_2_VERSION
 	case '2':
 		return TLS1_2_VERSION;
+#endif
 #ifdef TLS1_3_VERSION
-	/*
-	 * libressl uses version numbers starting with major version 2
-	 * but does not yet support TLS 1.3
-	 */
 	case '3':
 		return TLS1_3_VERSION;
-#endif
 #endif
 	default:
 		return -1;
@@ -158,7 +175,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 	int ret = ERR_CFG_UNKNOWN;
 	FILE *file;
 	struct stat stat;
-	char *buffer, *line;
+	char *buffer, *line, *saveptr = NULL;
 
 	file = fopen(filename, "r");
 	if (file == NULL)
@@ -188,8 +205,8 @@ int load_config(struct vpn_config *cfg, const char *filename)
 	buffer[stat.st_size] = '\0';
 
 	// Read line by line
-	for (line = strtok(buffer, "\n"); line != NULL;
-	     line = strtok(NULL, "\n")) {
+	for (line = strtok_r(buffer, "\n", &saveptr); line != NULL;
+	     line = strtok_r(NULL, "\n", &saveptr)) {
 		char *key, *equals, *val;
 		int i;
 
@@ -230,6 +247,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->gateway_host[FIELD_SIZE] = '\0';
 		} else if (strcmp(key, "port") == 0) {
 			unsigned long port = strtoul(val, NULL, 0);
+
 			if (port == 0 || port > 65535) {
 				log_warn("Bad port in config file: \"%lu\".\n",
 				         port);
@@ -249,6 +267,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->otp_prompt = strdup(val);
 		} else if (strcmp(key, "otp-delay") == 0) {
 			long otp_delay = strtol(val, NULL, 0);
+
 			if (otp_delay < 0 || otp_delay > UINT_MAX) {
 				log_warn("Bad value for otp-delay in config file: \"%s\".\n",
 				         val);
@@ -263,6 +282,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->realm[FIELD_SIZE] = '\0';
 		} else if (strcmp(key, "set-dns") == 0) {
 			int set_dns = strtob(val);
+
 			if (set_dns < 0) {
 				log_warn("Bad set-dns in config file: \"%s\".\n",
 				         val);
@@ -271,6 +291,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->set_dns = set_dns;
 		} else if (strcmp(key, "set-routes") == 0) {
 			int set_routes = strtob(val);
+
 			if (set_routes < 0) {
 				log_warn("Bad set-routes in config file: \"%s\".\n",
 				         val);
@@ -279,6 +300,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->set_routes = set_routes;
 		} else if (strcmp(key, "half-internet-routes") == 0) {
 			int half_internet_routes = strtob(val);
+
 			if (half_internet_routes < 0) {
 				log_warn("Bad half-internet-routes in config file: \"%s\".\n",
 				         val);
@@ -287,6 +309,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->half_internet_routes = half_internet_routes;
 		} else if (strcmp(key, "persistent") == 0) {
 			unsigned long persistent = strtoul(val, NULL, 0);
+
 			if (persistent > UINT_MAX) {
 				log_warn("Bad value for persistent in config file: \"%s\".\n",
 				         val);
@@ -296,6 +319,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 #if HAVE_USR_SBIN_PPPD
 		} else if (strcmp(key, "pppd-use-peerdns") == 0) {
 			int pppd_use_peerdns = strtob(val);
+
 			if (pppd_use_peerdns < 0) {
 				log_warn("Bad pppd-use-peerdns in config file: \"%s\".\n",
 				         val);
@@ -326,9 +350,23 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->ppp_system = strdup(val);
 #else
 			log_warn("Ignoring option \"%s\".\n", key);
+#endif
+		} else if (strcmp(key, "use-resolvconf") == 0) {
+#if HAVE_RESOLVCONF
+			int use_resolvconf = strtob(val);
+
+			if (use_resolvconf < 0) {
+				log_warn("Bad use-resolvconf value in config file: \"%s\".\n",
+				         val);
+				continue;
+			}
+			cfg->use_resolvconf = use_resolvconf;
+#else
+			log_warn("Ignoring option \"%s\".\n", key);
 #endif
 		} else if (strcmp(key, "use-syslog") == 0) {
 			int use_syslog = strtob(val);
+
 			if (use_syslog < 0) {
 				log_warn("Bad use-syslog in config file: \"%s\".\n",
 				         val);
@@ -357,6 +395,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			cfg->user_key = strdup(val);
 		} else if (strcmp(key, "insecure-ssl") == 0) {
 			int insecure_ssl = strtob(val);
+
 			if (insecure_ssl < 0) {
 				log_warn("Bad insecure-ssl in config file: \"%s\".\n",
 				         val);
@@ -369,6 +408,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L
 		} else if (strcmp(key, "min-tls") == 0) {
 			int min_tls = parse_min_tls(val);
+
 			if (min_tls == -1) {
 				log_warn("Bad min-tls in config file: \"%s\".\n",
 				         val);
@@ -379,6 +419,7 @@ int load_config(struct vpn_config *cfg, const char *filename)
 #endif
 		} else if (strcmp(key, "seclevel-1") == 0) {
 			int seclevel_1 = strtob(val);
+
 			if (seclevel_1 < 0) {
 				log_warn("Bad seclevel-1 in config file: \"%s\".\n",
 				         val);
@@ -396,7 +437,8 @@ int load_config(struct vpn_config *cfg, const char *filename)
 err_free:
 	free(buffer);
 err_close:
-	fclose(file);
+	if (fclose(file))
+		log_warn("Could not close %s (%s).\n", filename, strerror(errno));
 
 	return ret;
 }
@@ -422,6 +464,7 @@ void destroy_vpn_config(struct vpn_config *cfg)
 	free(cfg->cipher_list);
 	while (cfg->cert_whitelist != NULL) {
 		struct x509_digest *tmp = cfg->cert_whitelist->next;
+
 		free(cfg->cert_whitelist);
 		cfg->cert_whitelist = tmp;
 	}
@@ -453,6 +496,10 @@ void merge_config(struct vpn_config *dst, struct vpn_config *src)
 		dst->set_dns = src->set_dns;
 	if (src->pppd_use_peerdns != invalid_cfg.pppd_use_peerdns)
 		dst->pppd_use_peerdns = src->pppd_use_peerdns;
+#if HAVE_RESOLVCONF
+	if (src->use_resolvconf != invalid_cfg.use_resolvconf)
+		dst->use_resolvconf = src->use_resolvconf;
+#endif
 	if (src->use_syslog != invalid_cfg.use_syslog)
 		dst->use_syslog = src->use_syslog;
 	if (src->half_internet_routes != invalid_cfg.half_internet_routes)
@@ -514,6 +561,7 @@ void merge_config(struct vpn_config *dst, struct vpn_config *src)
 	if (src->cert_whitelist) {
 		while (dst->cert_whitelist != NULL) {
 			struct x509_digest *tmp = dst->cert_whitelist->next;
+
 			free(dst->cert_whitelist);
 			dst->cert_whitelist = tmp;
 		}
