@@ -32,12 +32,13 @@
 #include "tunnel.h"
 #include "log.h"
 
-#include <signal.h>
+#include <unistd.h>
 #include <arpa/inet.h>
 #include <netinet/tcp.h>
-#include <unistd.h>
-#include <string.h>
+
 #include <errno.h>
+#include <signal.h>
+#include <string.h>
 
 #if HAVE_MACH_MACH_H
 /* this is typical for mach kernel used on Mac OS X */
@@ -84,6 +85,7 @@ static unsigned long thread_id(void)
 static void init_ssl_locks(void)
 {
 	int i;
+
 	lockarray = (pthread_mutex_t *) OPENSSL_malloc(CRYPTO_num_locks() *
 	                sizeof(pthread_mutex_t));
 	for (i = 0; i < CRYPTO_num_locks(); i++)
@@ -94,6 +96,7 @@ static void init_ssl_locks(void)
 static void destroy_ssl_locks(void)
 {
 	int i;
+
 	CRYPTO_set_locking_callback(NULL);
 	for (i = 0; i < CRYPTO_num_locks(); i++)
 		pthread_mutex_destroy(&(lockarray[i]));
@@ -389,7 +392,7 @@ static inline void set_tunnel_ips(struct tunnel *tunnel,
 }
 
 #define printable_char(c) \
-    (c == '\t' || c == '\n' || (c >= ' ' && c <= '~'))
+	(c == '\t' || c == '\n' || (c >= ' ' && c <= '~'))
 
 static void debug_bad_packet(struct tunnel *tunnel, uint8_t *header)
 {
@@ -430,13 +433,26 @@ static void *ssl_read(void *arg)
 		struct ppp_packet *packet;
 		int ret;
 		uint8_t header[6];
+		static const char http_header[6] = "HTTP/1";
 		uint16_t total, magic, size;
 
 		ret = safe_ssl_read_all(tunnel->ssl_handle, header, 6);
 		if (ret < 0) {
 			log_debug("Error reading from SSL connection (%s).\n",
 			          err_ssl_str(ret));
-			goto exit;
+			break;
+		}
+
+		if (memcmp(header, http_header, 6) == 0) {
+			/*
+			 * When the SSL-VPN portal has not been set up to allow
+			 * tunnel mode for VPN clients, while it allows web mode
+			 * for web browsers, it returns an HTTP error instead of
+			 * a PPP packet:
+			 * HTTP/1.1 403 Forbidden
+			 */
+			log_error("Could not authenticate to the gateway. Please make sure tunnel mode is allowed by the gateway, check the realm, etc.\n");
+			break;
 		}
 
 		total = (header[0] << 8) | header[1];
@@ -462,7 +478,7 @@ static void *ssl_read(void *arg)
 			log_debug("Error reading from SSL connection (%s).\n",
 			          err_ssl_str(ret));
 			free(packet);
-			goto exit;
+			break;
 		}
 
 		log_debug("gateway ---> %s (%lu bytes)\n", PPP_DAEMON, packet->len);
@@ -471,7 +487,8 @@ static void *ssl_read(void *arg)
 
 		if (tunnel->state == STATE_CONNECTING) {
 			if (packet_is_ip_plus_dns(packet)) {
-				char line[57]; // 1 + 15 + 7 + 15 + 2 + 15 + 1 + 1
+				char line[ARRAY_SIZE("[xxx.xxx.xxx.xxx], ns [xxx.xxx.xxx.xxx, xxx.xxx.xxx.xxx], ns_suffix []") + MAX_DOMAIN_LENGTH];
+
 				set_tunnel_ips(tunnel, packet);
 				strcpy(line, "[");
 				strncat(line, inet_ntoa(tunnel->ipv4.ip_addr), 15);
@@ -479,17 +496,21 @@ static void *ssl_read(void *arg)
 				strncat(line, inet_ntoa(tunnel->ipv4.ns1_addr), 15);
 				strcat(line, ", ");
 				strncat(line, inet_ntoa(tunnel->ipv4.ns2_addr), 15);
+				if (tunnel->ipv4.dns_suffix) {
+					strcat(line, "], ns_suffix [");
+					strncat(line, tunnel->ipv4.dns_suffix,
+					        MAX_DOMAIN_LENGTH);
+				}
 				strcat(line, "]");
 				log_info("Got addresses: %s\n", line);
 			}
 			if (packet_is_end_negociation(packet)) {
-				log_info("negotiation complete\n");
+				log_info("Negotiation complete.\n");
 				SEM_POST(&sem_if_config);
 			}
 		}
 	}
 
-exit:
 	// Send message to main thread to stop other threads
 	SEM_POST(&sem_stop_io);
 	return NULL;
@@ -586,6 +607,8 @@ static void sig_handler(int signo)
 int io_loop(struct tunnel *tunnel)
 {
 	int tcp_nodelay_flag = 1;
+	int ret = 0;		// keep track of pthread_* return value
+	int fatal = 0;		// indicate a fatal error during pthread_* calls
 
 	pthread_t pty_read_thread;
 	pthread_t pty_write_thread;
@@ -602,6 +625,8 @@ int io_loop(struct tunnel *tunnel)
 
 	init_ssl_locks();
 
+	init_hdlc();
+
 	/*
 	 * I noticed that using TCP_NODELAY (i.e. disabling Nagle's algorithm)
 	 * gives much better performance. Probably because setting up the VPN
@@ -623,6 +648,7 @@ int io_loop(struct tunnel *tunnel)
 #if !HAVE_MACH_MACH_H
 	// Disable SIGINT and SIGTERM for the future spawned threads
 	sigset_t sigset, oldset;
+
 	sigemptyset(&sigset);
 	sigaddset(&sigset, SIGINT);
 	sigaddset(&sigset, SIGTERM);
@@ -638,16 +664,36 @@ int io_loop(struct tunnel *tunnel)
 	if (signal(SIGHUP, SIG_IGN) == SIG_ERR)
 		goto err_signal;
 
-	if (pthread_create(&pty_read_thread, NULL, pppd_read, tunnel))
+	// create all workers, stop on first error and bail out
+	ret = pthread_create(&pty_read_thread, NULL, pppd_read, tunnel);
+	if (ret != 0) {
+		log_debug("Error creating pty_read_thread: %s\n", strerror(ret));
 		goto err_thread;
-	if (pthread_create(&pty_write_thread, NULL, pppd_write, tunnel))
+	}
+
+	ret = pthread_create(&pty_write_thread, NULL, pppd_write, tunnel);
+	if (ret != 0) {
+		log_debug("Error creating pty_write_thread: %s\n", strerror(ret));
 		goto err_thread;
-	if (pthread_create(&ssl_read_thread, NULL, ssl_read, tunnel))
+	}
+
+	ret = pthread_create(&ssl_read_thread, NULL, ssl_read, tunnel);
+	if (ret != 0) {
+		log_debug("Error creating ssl_read_thread: %s\n", strerror(ret));
 		goto err_thread;
-	if (pthread_create(&ssl_write_thread, NULL, ssl_write, tunnel))
+	}
+
+	ret = pthread_create(&ssl_write_thread, NULL, ssl_write, tunnel);
+	if (ret != 0) {
+		log_debug("Error creating ssl_write_thread: %s\n", strerror(ret));
 		goto err_thread;
-	if (pthread_create(&if_config_thread, NULL, if_config, tunnel))
+	}
+
+	ret = pthread_create(&if_config_thread, NULL, if_config, tunnel);
+	if (ret != 0) {
+		log_debug("Error creating if_config_thread: %s\n", strerror(ret));
 		goto err_thread;
+	}
 
 #if !HAVE_MACH_MACH_H
 	// Restore the signal for the main thread
@@ -658,17 +704,58 @@ int io_loop(struct tunnel *tunnel)
 	SEM_WAIT(&sem_stop_io);
 
 	log_info("Cancelling threads...\n");
-	pthread_cancel(if_config_thread);
-	pthread_cancel(ssl_write_thread);
-	pthread_cancel(ssl_read_thread);
-	pthread_cancel(pty_write_thread);
-	pthread_cancel(pty_read_thread);
+	// no goto err_thread here, try to cancel all threads
+	ret = pthread_cancel(if_config_thread);
+	if (ret != 0)
+		log_debug("Error canceling if_config_thread: %s\n", strerror(ret));
 
-	pthread_join(if_config_thread, NULL);
-	pthread_join(ssl_write_thread, NULL);
-	pthread_join(ssl_read_thread, NULL);
-	pthread_join(pty_write_thread, NULL);
-	pthread_join(pty_read_thread, NULL);
+	ret = pthread_cancel(ssl_write_thread);
+	if (ret != 0)
+		log_debug("Error canceling ssl_write_thread: %s\n", strerror(ret));
+
+	ret = pthread_cancel(ssl_read_thread);
+	if (ret != 0)
+		log_debug("Error canceling safe_ssl_read_thread: %s\n", strerror(ret));
+
+	ret = pthread_cancel(pty_write_thread);
+	if (ret != 0)
+		log_debug("Error canceling pty_write_thread: %s\n", strerror(ret));
+
+	ret = pthread_cancel(pty_read_thread);
+	if (ret != 0)
+		log_debug("Error canceling pty_read_thread: %s\n", strerror(ret));
+
+	log_info("Cleanup, joining threads...\n");
+	// failure to clean is a possible zombie thread, consider it fatal
+	ret = pthread_join(if_config_thread, NULL);
+	if (ret != 0) {
+		log_debug("Error joining if_config_thread: %s\n", strerror(ret));
+		fatal = 1;
+	}
+
+	ret = pthread_join(ssl_write_thread, NULL);
+	if (ret != 0) {
+		log_debug("Error joining ssl_write_thread: %s\n", strerror(ret));
+		fatal = 1;
+	}
+
+	ret = pthread_join(ssl_read_thread, NULL);
+	if (ret != 0) {
+		log_debug("Error joining ssl_read_thread: %s\n", strerror(ret));
+		fatal = 1;
+	}
+
+	ret = pthread_join(pty_write_thread, NULL);
+	if (ret != 0) {
+		log_debug("Error joining pty_write_thread: %s\n", strerror(ret));
+		fatal = 1;
+	}
+
+	ret = pthread_join(pty_read_thread, NULL);
+	if (ret != 0) {
+		log_debug("Error joining pty_read_thread: %s\n", strerror(ret));
+		fatal = 1;
+	}
 
 	destroy_ssl_locks();
 
@@ -679,6 +766,10 @@ int io_loop(struct tunnel *tunnel)
 	SEM_DESTROY(&sem_if_config);
 	SEM_DESTROY(&sem_pppd_ready);
 
+	// should we have detected a fatal error
+	if (fatal)
+		goto err_thread;
+
 	return 0;
 
 err_thread:
