@@ -20,16 +20,21 @@
 #include "userinput.h"
 #include "log.h"
 
-#include <getopt.h>
+#include <openssl/ssl.h>
+
 #include <unistd.h>
+#include <getopt.h>
+
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <limits.h>
 
 #define PWD_BUFSIZ	4096
 
-#if HAVE_USR_SBIN_PPPD
+#if HAVE_USR_SBIN_PPPD && HAVE_USR_SBIN_PPP
+#error "Both HAVE_USR_SBIN_PPPD and HAVE_USR_SBIN_PPP have been defined."
+#elif HAVE_USR_SBIN_PPPD
 #define PPPD_USAGE \
 "                    [--pppd-use-peerdns=<0|1>] [--pppd-log=<file>]\n" \
 "                    [--pppd-ifname=<string>] [--pppd-ipparam=<string>]\n" \
@@ -45,18 +50,29 @@
 "  --pppd-plugin=<file>          Use specified pppd plugin instead of configuring\n" \
 "                                resolver and routes directly.\n" \
 "  --pppd-ifname=<string>        Set the pppd interface name, if supported by pppd.\n" \
-"  --pppd-ipparam=<string>       Provides  an extra parameter to the ip-up, ip-pre-up\n" \
+"  --pppd-ipparam=<string>       Provides an extra parameter to the ip-up, ip-pre-up\n" \
 "                                and ip-down scripts. See man (8) pppd\n" \
 "  --pppd-call=<name>            Move most pppd options from pppd cmdline to\n" \
 "                                /etc/ppp/peers/<name> and invoke pppd with\n" \
 "                                'call <name>'\n"
-#endif
-#if HAVE_USR_SBIN_PPP
+#elif HAVE_USR_SBIN_PPP
 #define PPPD_USAGE \
 "                    [--ppp-system=<system>]\n"
 #define PPPD_HELP \
-"  --ppp-system=<system>         connect to the specified system as defined in\n" \
+"  --ppp-system=<system>         Connect to the specified system as defined in\n" \
 "                                /etc/ppp/ppp.conf\n"
+#else
+#error "Neither HAVE_USR_SBIN_PPPD nor HAVE_USR_SBIN_PPP have been defined."
+#endif
+
+#if HAVE_RESOLVCONF
+#define RESOLVCONF_USAGE \
+"[--use-resolvconf=<0|1>] "
+#define RESOLVCONF_HELP \
+"  --use-resolvconf=[01]         If possible use resolvconf to update /etc/resolv.conf\n"
+#else
+#define RESOLVCONF_USAGE ""
+#define RESOLVCONF_HELP ""
 #endif
 
 #define usage \
@@ -66,7 +82,7 @@
 "                    [--otp-prompt=<prompt>] [--set-routes=<0|1>]\n" \
 "                    [--half-internet-routes=<0|1>] [--set-dns=<0|1>]\n" \
 PPPD_USAGE \
-"                    [--ca-file=<file>]\n" \
+"                    " RESOLVCONF_USAGE "[--ca-file=<file>]\n" \
 "                    [--user-cert=<file>] [--user-key=<file>]\n" \
 "                    [--trusted-cert=<digest>] [--use-syslog]\n" \
 "                    [--persistent=<interval>] [-c <file>] [-v|-q]\n" \
@@ -81,15 +97,16 @@ PPPD_USAGE \
 "the gateway and this process.\n" \
 "\n"
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-#define help_min_tls ""
+#ifdef TLS1_3_VERSION
+#define help_cipher_list " Applies to TLS v1.2 or\n" \
+"                                lower only, not to be used with TLS v1.3 ciphers."
+#define help_seclevel_1 " Applies to TLS v1.2 or lower only."
 #else
-#define help_min_tls " This option\n" \
-"                                is not supported by your OpenSSL library."
+#define help_cipher_list ""
+#define help_seclevel_1 ""
 #endif
 
-
-#define help_options \
+#define help_options_part1 \
 "Options:\n" \
 "  -h --help                     Show this help message and exit.\n" \
 "  --version                     Show version and exit.\n" \
@@ -125,7 +142,9 @@ PPPD_USAGE \
 "                                certificate will be matched against this value.\n" \
 "                                <digest> is the X509 certificate's sha256 sum.\n" \
 "                                This option can be used multiple times to trust\n" \
-"                                several certificates.\n" \
+"                                several certificates.\n"
+
+#define help_options_part2 \
 "  --insecure-ssl                Do not disable insecure SSL protocols/ciphers.\n" \
 "                                Also enable TLS v1.0 if applicable.\n" \
 "                                If your server requires a specific cipher or protocol,\n" \
@@ -133,12 +152,12 @@ PPPD_USAGE \
 "  --cipher-list=<ciphers>       OpenSSL ciphers to use. If default does not work\n" \
 "                                you can try with the cipher suggested in the output\n" \
 "                                of 'openssl s_client -connect <host:port>'\n" \
-"                                (e.g. AES256-GCM-SHA384)\n" \
+"                                (e.g. AES256-GCM-SHA384)." help_cipher_list "\n" \
 "  --min-tls                     Use minimum TLS version instead of system default.\n" \
-"                                Valid values are 1.0, 1.1, 1.2, 1.3." help_min_tls "\n" \
+"                                Valid values are 1.0, 1.1, 1.2, 1.3.\n" \
 "  --seclevel-1                  If --cipher-list is not specified, add @SECLEVEL=1 to\n" \
 "                                (compiled in) list of ciphers. This lowers limits on\n" \
-"                                dh key.\n" \
+"                                dh key." help_seclevel_1 "\n" \
 "  --persistent=<interval>       Run the vpn persistently in a loop and try to re-\n" \
 "                                connect every <interval> seconds when dropping out\n" \
 "  -v                            Increase verbosity. Can be used multiple times\n" \
@@ -146,7 +165,6 @@ PPPD_USAGE \
 "  -q                            Decrease verbosity. Can be used multiple times\n" \
 "                                to be even less verbose.\n"
 
-
 #define help_config \
 "\n" \
 "Config file:\n" \
@@ -186,6 +204,9 @@ int main(int argc, char **argv)
 		.use_syslog = 0,
 		.half_internet_routes = 0,
 		.persistent = 0,
+#if HAVE_RESOLVCONF
+		.use_resolvconf = USE_RESOLVCONF,
+#endif
 #if HAVE_USR_SBIN_PPPD
 		.pppd_use_peerdns = 0,
 		.pppd_log = NULL,
@@ -201,55 +222,61 @@ int main(int argc, char **argv)
 		.user_cert = NULL,
 		.user_key = NULL,
 		.insecure_ssl = 0,
+#ifdef TLS1_2_VERSION
+		.min_tls = TLS1_2_VERSION,
+#else
 		.min_tls = 0,
+#endif
 		.seclevel_1 = 0,
 		.cipher_list = NULL,
-		.cert_whitelist = NULL
+		.cert_whitelist = NULL,
+		.use_engine = 0
 	};
 	struct vpn_config cli_cfg = invalid_cfg;
 
 	const struct option long_options[] = {
-		{"help",            no_argument,       0, 'h'},
-		{"version",         no_argument,       0, 0},
-		{"config",          required_argument, 0, 'c'},
-		{"pinentry",        required_argument, 0, 0},
-		{"realm",           required_argument, 0, 0},
-		{"username",        required_argument, 0, 'u'},
-		{"password",        required_argument, 0, 'p'},
-		{"otp",             required_argument, 0, 'o'},
-		{"otp-prompt",      required_argument, 0, 0},
-		{"otp-delay",       required_argument, 0, 0},
-		{"set-routes",	    required_argument, 0, 0},
+		{"help",            no_argument,       NULL, 'h'},
+		{"version",         no_argument,       NULL, 0},
+		{"config",          required_argument, NULL, 'c'},
+		{"pinentry",        required_argument, NULL, 0},
+		{"realm",           required_argument, NULL, 0},
+		{"username",        required_argument, NULL, 'u'},
+		{"password",        required_argument, NULL, 'p'},
+		{"otp",             required_argument, NULL, 'o'},
+		{"otp-prompt",      required_argument, NULL, 0},
+		{"otp-delay",       required_argument, NULL, 0},
+		{"set-routes",	    required_argument, NULL, 0},
 		{"no-routes",       no_argument, &cli_cfg.set_routes, 0},
-		{"half-internet-routes", required_argument, 0, 0},
-		{"set-dns",         required_argument, 0, 0},
+		{"half-internet-routes", required_argument, NULL, 0},
+		{"set-dns",         required_argument, NULL, 0},
 		{"no-dns",          no_argument, &cli_cfg.set_dns, 0},
 		{"use-syslog",      no_argument, &cli_cfg.use_syslog, 1},
-		{"persistent",      required_argument, 0, 0},
-		{"ca-file",         required_argument, 0, 0},
-		{"user-cert",       required_argument, 0, 0},
-		{"user-key",        required_argument, 0, 0},
-		{"trusted-cert",    required_argument, 0, 0},
+		{"persistent",      required_argument, NULL, 0},
+		{"ca-file",         required_argument, NULL, 0},
+		{"user-cert",       required_argument, NULL, 0},
+		{"user-key",        required_argument, NULL, 0},
+		{"trusted-cert",    required_argument, NULL, 0},
 		{"insecure-ssl",    no_argument, &cli_cfg.insecure_ssl, 1},
-		{"cipher-list",     required_argument, 0, 0},
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-		{"min-tls",         required_argument, 0, 0},
-#endif
+		{"cipher-list",     required_argument, NULL, 0},
+		{"min-tls",         required_argument, NULL, 0},
 		{"seclevel-1",      no_argument, &cli_cfg.seclevel_1, 1},
 #if HAVE_USR_SBIN_PPPD
-		{"pppd-use-peerdns", required_argument, 0, 0},
+		{"pppd-use-peerdns", required_argument, NULL, 0},
 		{"pppd-no-peerdns", no_argument, &cli_cfg.pppd_use_peerdns, 0},
-		{"pppd-log",        required_argument, 0, 0},
-		{"pppd-plugin",     required_argument, 0, 0},
-		{"pppd-ipparam",    required_argument, 0, 0},
-		{"pppd-ifname",     required_argument, 0, 0},
-		{"pppd-call",       required_argument, 0, 0},
-		{"plugin",          required_argument, 0, 0}, // deprecated
+		{"pppd-log",        required_argument, NULL, 0},
+		{"pppd-plugin",     required_argument, NULL, 0},
+		{"pppd-ipparam",    required_argument, NULL, 0},
+		{"pppd-ifname",     required_argument, NULL, 0},
+		{"pppd-call",       required_argument, NULL, 0},
+		{"plugin",          required_argument, NULL, 0}, // deprecated
 #endif
 #if HAVE_USR_SBIN_PPP
-		{"ppp-system",      required_argument, 0, 0},
+		{"ppp-system",      required_argument, NULL, 0},
 #endif
-		{0, 0, 0, 0}
+#if HAVE_RESOLVCONF
+		{"use-resolvconf",  required_argument, NULL, 0},
+#endif
+		{NULL, 0, NULL, 0}
 	};
 
 	init_logging();
@@ -273,6 +300,8 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "version") == 0) {
 				printf(VERSION "\n");
+				if (strcmp(&REVISION[1], VERSION))
+					log_debug("revision " REVISION "\n");
 				ret = EXIT_SUCCESS;
 				goto exit;
 			}
@@ -280,6 +309,7 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "pppd-use-peerdns") == 0) {
 				int pppd_use_peerdns = strtob(optarg);
+
 				if (pppd_use_peerdns < 0) {
 					log_warn("Bad pppd-use-peerdns option: \"%s\"\n",
 					         optarg);
@@ -327,6 +357,20 @@ int main(int argc, char **argv)
 				cfg.ppp_system = strdup(optarg);
 				break;
 			}
+#endif
+#if HAVE_RESOLVCONF
+			if (strcmp(long_options[option_index].name,
+			           "use-resolvconf") == 0) {
+				int use_resolvconf = strtob(optarg);
+
+				if (use_resolvconf < 0) {
+					log_warn("Bad use-resolvconf option: \"%s\"\n",
+					         optarg);
+					break;
+				}
+				cli_cfg.use_resolvconf = use_resolvconf;
+				break;
+			}
 #endif
 			if (strcmp(long_options[option_index].name,
 			           "ca-file") == 0) {
@@ -365,10 +409,10 @@ int main(int argc, char **argv)
 				cli_cfg.cipher_list = strdup(optarg);
 				break;
 			}
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
 			if (strcmp(long_options[option_index].name,
 			           "min-tls") == 0) {
 				int min_tls = parse_min_tls(optarg);
+
 				if (min_tls == -1) {
 					log_warn("Bad min-tls option: \"%s\"\n",
 					         optarg);
@@ -377,7 +421,6 @@ int main(int argc, char **argv)
 				}
 				break;
 			}
-#endif
 			if (strcmp(long_options[option_index].name,
 			           "otp-prompt") == 0) {
 				cli_cfg.otp_prompt = strdup(optarg);
@@ -386,6 +429,7 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "set-routes") == 0) {
 				int set_routes = strtob(optarg);
+
 				if (set_routes < 0) {
 					log_warn("Bad set-routes option: \"%s\"\n",
 					         optarg);
@@ -397,6 +441,7 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "half-internet-routes") == 0) {
 				int half_internet_routes = strtob(optarg);
+
 				if (half_internet_routes < 0) {
 					log_warn("Bad half-internet-routes option: \"%s\"\n",
 					         optarg);
@@ -408,6 +453,7 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "otp-delay") == 0) {
 				long otp_delay = strtol(optarg, NULL, 0);
+
 				if (otp_delay < 0 || otp_delay > UINT_MAX) {
 					log_warn("Bad otp-delay option: \"%s\"\n",
 					         optarg);
@@ -419,6 +465,7 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "persistent") == 0) {
 				long persistent = strtol(optarg, NULL, 0);
+
 				if (persistent < 0 || persistent > UINT_MAX) {
 					log_warn("Bad persistent option: \"%s\"\n",
 					         optarg);
@@ -430,6 +477,7 @@ int main(int argc, char **argv)
 			if (strcmp(long_options[option_index].name,
 			           "set-dns") == 0) {
 				int set_dns = strtob(optarg);
+
 				if (set_dns < 0) {
 					log_warn("Bad set-dns option: \"%s\"\n", optarg);
 					break;
@@ -439,8 +487,9 @@ int main(int argc, char **argv)
 			}
 			goto user_error;
 		case 'h':
-			printf("%s%s%s%s%s", usage, summary, help_options,
-			       PPPD_HELP, help_config);
+			printf("%s%s%s%s%s%s%s", usage, summary,
+			       help_options_part1, help_options_part2,
+			       PPPD_HELP, RESOLVCONF_HELP, help_config);
 			ret = EXIT_SUCCESS;
 			goto exit;
 		case 'v':
@@ -477,6 +526,8 @@ int main(int argc, char **argv)
 	log_debug_all("ATTENTION: the output contains sensitive information such as the THE CLEAR TEXT PASSWORD.\n");
 
 	log_debug("openfortivpn " VERSION "\n");
+	if (strcmp(&REVISION[1], VERSION))
+		log_debug("revision " REVISION "\n");
 
 	// Load config file
 	if (config_file[0] != '\0') {
@@ -529,6 +580,12 @@ int main(int argc, char **argv)
 	// If username but no password given, interactively ask user
 	if (cfg.password == NULL && cfg.username[0] != '\0') {
 		char *tmp_password = malloc(PWD_BUFSIZ); // allocate large buffer
+
+		if (tmp_password == NULL) {
+			log_error("malloc: %s\n", strerror(errno));
+			ret = EXIT_FAILURE;
+			goto exit;
+		}
 		read_password(cfg.pinentry, "password",
 		              "VPN account password: ", tmp_password, PWD_BUFSIZ);
 		cfg.password = strdup(tmp_password); // copy string of correct size
