@@ -33,35 +33,38 @@
 #include "openssl_hostname_validation.h"
 #endif
 
+#include <openssl/err.h>
+#include <openssl/engine.h>
+#include <openssl/ui.h>
+#include <openssl/x509v3.h>
+#if HAVE_SYSTEMD
+#include <systemd/sd-daemon.h>
+#endif
+
 #include <unistd.h>
+#include <arpa/inet.h>
 #include <fcntl.h>
 #include <ifaddrs.h>
-#include <netdb.h>
-#include <errno.h>
-#include <string.h>
 #include <net/if.h>
-#include <arpa/inet.h>
-#include <sys/socket.h>
-#include <sys/ioctl.h>
-#include <openssl/err.h>
-#include <openssl/x509v3.h>
-#include <openssl/engine.h>
+#include <netdb.h>
 #if HAVE_PTY_H
 #include <pty.h>
 #elif HAVE_UTIL_H
 #include <util.h>
 #endif
-#include <termios.h>
-#include <signal.h>
+#include <sys/types.h>
+#include <sys/socket.h>
 #include <sys/wait.h>
-#if HAVE_SYSTEMD
-#include <systemd/sd-daemon.h>
+#include <sys/ioctl.h>
+#include <termios.h>
+#if HAVE_LIBUTIL_H
+#include <libutil.h>
 #endif
 
-// we use this constant in the source, so define a fallback if not defined
-#ifndef OPENSSL_API_COMPAT
-#define OPENSSL_API_COMPAT 0x0908000L
-#endif
+#include <errno.h>
+#include <signal.h>
+#include <string.h>
+
 
 struct ofv_varr {
 	unsigned int cap;	// current capacity
@@ -74,6 +77,7 @@ static int ofv_append_varr(struct ofv_varr *p, const char *x)
 	if (p->off + 1 >= p->cap) {
 		const char **ndata;
 		unsigned int ncap = (p->off + 1) * 2;
+
 		if (p->off + 1 >= ncap) {
 			log_error("%s: ncap exceeded\n", __func__);
 			return 1;
@@ -131,7 +135,11 @@ static int on_ppp_if_up(struct tunnel *tunnel)
 
 static int on_ppp_if_down(struct tunnel *tunnel)
 {
-	log_info("Setting ppp interface down.\n");
+#if HAVE_SYSTEMD
+	sd_notify(0, "STOPPING=1");
+#endif
+
+	log_info("Setting %s interface down.\n", tunnel->ppp_iface);
 
 	if (tunnel->config->set_routes) {
 		log_info("Restoring routes...\n");
@@ -161,6 +169,7 @@ static int pppd_run(struct tunnel *tunnel)
 #endif
 
 	static const char ppp_path[] = PPP_PATH;
+
 	if (access(ppp_path, F_OK) != 0) {
 		log_error("%s: %s.\n", ppp_path, strerror(errno));
 		return 1;
@@ -170,7 +179,7 @@ static int pppd_run(struct tunnel *tunnel)
 	slave_stderr = dup(STDERR_FILENO);
 
 	if (slave_stderr < 0) {
-		log_error("slave stderr %s\n", strerror(errno));
+		log_error("slave stderr: %s\n", strerror(errno));
 		return 1;
 	}
 
@@ -185,7 +194,8 @@ static int pppd_run(struct tunnel *tunnel)
 		struct ofv_varr pppd_args = { 0, 0, NULL };
 
 		dup2(slave_stderr, STDERR_FILENO);
-		close(slave_stderr);
+		if (close(slave_stderr))
+			log_warn("Could not close slave stderr (%s).\n", strerror(errno));
 
 #if HAVE_USR_SBIN_PPP
 		/*
@@ -199,17 +209,25 @@ static int pppd_run(struct tunnel *tunnel)
 			"-direct"
 		};
 		for (unsigned int i = 0; i < ARRAY_SIZE(v); i++)
-			if (ofv_append_varr(&pppd_args, v[i]))
+			if (ofv_append_varr(&pppd_args, v[i])) {
+				free(pppd_args.data);
 				return 1;
+			}
 #endif
 #if HAVE_USR_SBIN_PPPD
 		if (tunnel->config->pppd_call) {
-			if (ofv_append_varr(&pppd_args, ppp_path))
+			if (ofv_append_varr(&pppd_args, ppp_path)) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, "call"))
+			}
+			if (ofv_append_varr(&pppd_args, "call")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_call))
+			}
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_call)) {
+				free(pppd_args.data);
 				return 1;
+			}
 		} else {
 			static const char *const v[] = {
 				ppp_path,
@@ -227,63 +245,94 @@ static int pppd_run(struct tunnel *tunnel)
 				"mru", "1354"
 			};
 			for (unsigned int i = 0; i < ARRAY_SIZE(v); i++)
-				if (ofv_append_varr(&pppd_args, v[i]))
+				if (ofv_append_varr(&pppd_args, v[i])) {
+					free(pppd_args.data);
 					return 1;
+				}
 		}
 		if (tunnel->config->pppd_use_peerdns)
-			if (ofv_append_varr(&pppd_args, "usepeerdns"))
+			if (ofv_append_varr(&pppd_args, "usepeerdns")) {
+				free(pppd_args.data);
 				return 1;
+			}
 		if (tunnel->config->pppd_log) {
-			if (ofv_append_varr(&pppd_args, "debug"))
+			if (ofv_append_varr(&pppd_args, "debug")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, "logfile"))
+			}
+			if (ofv_append_varr(&pppd_args, "logfile")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_log))
+			}
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_log)) {
+				free(pppd_args.data);
 				return 1;
+			}
 		} else {
 			/*
 			 * pppd defaults to logging to fd=1, clobbering the
 			 * actual PPP data
 			 */
-			if (ofv_append_varr(&pppd_args, "logfd"))
+			if (ofv_append_varr(&pppd_args, "logfd")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, "2"))
+			}
+			if (ofv_append_varr(&pppd_args, "2")) {
+				free(pppd_args.data);
 				return 1;
+			}
 		}
 		if (tunnel->config->pppd_plugin) {
-			if (ofv_append_varr(&pppd_args, "plugin"))
+			if (ofv_append_varr(&pppd_args, "plugin")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_plugin))
+			}
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_plugin)) {
+				free(pppd_args.data);
 				return 1;
+			}
 		}
 		if (tunnel->config->pppd_ipparam) {
-			if (ofv_append_varr(&pppd_args, "ipparam"))
+			if (ofv_append_varr(&pppd_args, "ipparam")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_ipparam))
+			}
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_ipparam)) {
+				free(pppd_args.data);
 				return 1;
+			}
 		}
 		if (tunnel->config->pppd_ifname) {
-			if (ofv_append_varr(&pppd_args, "ifname"))
+			if (ofv_append_varr(&pppd_args, "ifname")) {
+				free(pppd_args.data);
 				return 1;
-			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_ifname))
+			}
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_ifname)) {
+				free(pppd_args.data);
 				return 1;
+			}
 		}
 #endif
 #if HAVE_USR_SBIN_PPP
 		if (tunnel->config->ppp_system) {
-			if (ofv_append_varr(&pppd_args, tunnel->config->ppp_system))
+			if (ofv_append_varr(&pppd_args, tunnel->config->ppp_system)) {
+				free(pppd_args.data);
 				return 1;
+			}
 		}
 #endif
 
-		close(tunnel->ssl_socket);
+		if (close(tunnel->ssl_socket))
+			log_warn("Could not close ssl socket (%s).\n", strerror(errno));
 		execv(pppd_args.data[0], (char *const *)pppd_args.data);
 		free(pppd_args.data);
 
-		fprintf(stderr, "execvp: %s\n", strerror(errno));
+		fprintf(stderr, "execv: %s\n", strerror(errno));
 		_exit(EXIT_FAILURE);
 	} else {
-		close(slave_stderr);
+		if (close(slave_stderr))
+			log_error("Could not close slave stderr (%s).\n",
+			          strerror(errno));
 		if (pid == -1) {
 			log_error("forkpty: %s\n", strerror(errno));
 			return 1;
@@ -292,6 +341,7 @@ static int pppd_run(struct tunnel *tunnel)
 
 	// Set non-blocking
 	int flags = fcntl(amaster, F_GETFL, 0);
+
 	if (flags == -1)
 		flags = 0;
 	if (fcntl(amaster, F_SETFL, flags | O_NONBLOCK) == -1) {
@@ -330,17 +380,20 @@ static const char * const pppd_message[] = {
 
 static int pppd_terminate(struct tunnel *tunnel)
 {
-	close(tunnel->pppd_pty);
+	if (close(tunnel->pppd_pty))
+		log_warn("Could not close pppd pty (%s).\n", strerror(errno));
 
 	log_debug("Waiting for %s to exit...\n", PPP_DAEMON);
 
 	int status;
+
 	if (waitpid(tunnel->pppd_pid, &status, 0) == -1) {
 		log_error("waitpid: %s\n", strerror(errno));
 		return 1;
 	}
 	if (WIFEXITED(status)) {
 		int exit_status = WEXITSTATUS(status);
+
 		log_debug("waitpid: %s exit status code %d\n",
 		          PPP_DAEMON, exit_status);
 #if HAVE_USR_SBIN_PPPD
@@ -383,6 +436,7 @@ static int pppd_terminate(struct tunnel *tunnel)
 #endif
 	} else if (WIFSIGNALED(status)) {
 		int signal_number = WTERMSIG(status);
+
 		log_debug("waitpid: %s terminated by signal %d\n",
 		          PPP_DAEMON, signal_number);
 		log_error("%s: terminated by signal: %s\n",
@@ -500,7 +554,7 @@ static int tcp_connect(struct tunnel *tunnel)
 		else
 			proxy_host += 3;
 		// split host and port
-		proxy_port = index(proxy_host, ':');
+		proxy_port = strchr(proxy_host, ':');
 		if (proxy_port != NULL) {
 			proxy_port[0] = '\0';
 			proxy_port++;
@@ -559,7 +613,9 @@ static int tcp_connect(struct tunnel *tunnel)
 		        tunnel->config->gateway_port,
 		        inet_ntoa(tunnel->config->gateway_ip),
 		        tunnel->config->gateway_port);
+
 		ssize_t bytes_written = write(handle, request, strlen(request));
+
 		if (bytes_written != strlen(request)) {
 			log_error("write error while talking to proxy: %s\n",
 			          strerror(errno));
@@ -572,62 +628,57 @@ static int tcp_connect(struct tunnel *tunnel)
 
 		memset(&(request), '\0', sizeof(request));
 		for (int j = 0; response == NULL; j++) {
-			/*
-			 * Coverity detected a defect:
-			 *  CID 200508: String not null terminated (STRING_NULL)
-			 *
-			 * It is actually a false positive:
-			 * • Function memset() initializes 'request' with '\0'
-			 * • Function read() gets a single char into: request[j]
-			 * • The final '\0' cannot be overwritten because:
-			 *   	j < ARRAY_SIZE(request) - 1
-			 */
-			ssize_t bytes_read = read(handle, &(request[j]), 1);
-			if (bytes_read < 1) {
+			if (j >= ARRAY_SIZE(request) - 1) {
 				log_error("Proxy response is unexpectedly large and cannot fit in the %lu-bytes buffer.\n",
 				          ARRAY_SIZE(request));
 				goto err_proxy_response;
 			}
 
-			// detect "200"
+			ssize_t bytes_read = read(handle, &(request[j]), 1);
 			static const char HTTP_STATUS_200[] = "200";
+
+			// we have reached the end of the data sent by the proxy
+			// and have not seen HTTP status code 200
+			if (bytes_read < 1) {
+				log_error("Proxy response does not contain \"%s\" as expected.\n",
+				          HTTP_STATUS_200);
+				goto err_proxy_response;
+			}
+
+			// detect "200"
 			response = strstr(request, HTTP_STATUS_200);
 
 			// detect end-of-line after "200"
 			if (response != NULL) {
 				/*
-				 * RFC2616 states in section 2.2 Basic Rules:
-				 * 	CR     = <US-ASCII CR, carriage return (13)>
-				 * 	LF     = <US-ASCII LF, linefeed (10)>
-				 * 	HTTP/1.1 defines the sequence CR LF as the
-				 * 	end-of-line marker for all protocol elements
-				 * 	except the entity-body (see appendix 19.3
-				 * 	for tolerant applications).
-				 * 		CRLF   = CR LF
+				 * RFC 2616 states in section 2.2 Basic Rules:
+				 *      CR     = <US-ASCII CR, carriage return (13)>
+				 *      LF     = <US-ASCII LF, linefeed (10)>
+				 *      HTTP/1.1 defines the sequence CR LF as the
+				 *      end-of-line marker for all protocol elements
+				 *      except the entity-body (see appendix 19.3
+				 *      for tolerant applications).
+				 *              CRLF   = CR LF
 				 *
-				 * RFC2616 states in section 19.3 Tolerant Applications:
-				 * 	The line terminator for message-header fields
-				 * 	is the sequence CRLF. However, we recommend
-				 * 	that applications, when parsing such headers,
-				 * 	recognize a single LF as a line terminator
-				 * 	and ignore the leading CR.
+				 * RFC 2616 states in section 19.3 Tolerant Applications:
+				 *      The line terminator for message-header fields
+				 *      is the sequence CRLF. However, we recommend
+				 *      that applications, when parsing such headers,
+				 *      recognize a single LF as a line terminator
+				 *      and ignore the leading CR.
 				 */
 				static const char *const HTTP_EOL[] = {
 					"\r\n\r\n",
 					"\n\n"
 				};
 				const char *eol = NULL;
+
 				for (int i = 0; (i < ARRAY_SIZE(HTTP_EOL)) &&
 				     (eol == NULL); i++)
 					eol = strstr(response, HTTP_EOL[i]);
 				response = eol;
 			}
 
-			if (j > ARRAY_SIZE(request) - 2) {
-				log_error("Proxy response does not contain \"%s\" as expected.\n",
-				          HTTP_STATUS_200);
-				goto err_proxy_response;
-			}
 		}
 
 		free(env_proxy); // release memory allocated by strdup()
@@ -639,7 +690,8 @@ err_proxy_response:
 err_connect:
 	free(env_proxy); // release memory allocated by strdup()
 err_strdup:
-	close(handle);
+	if (close(handle))
+		log_warn("Could not close socket (%s).\n", strerror(errno));
 err_socket:
 	return -1;
 }
@@ -655,10 +707,12 @@ static int ssl_verify_cert(struct tunnel *tunnel)
 	char *line;
 	int i;
 	X509_NAME *subj;
+	char *saveptr = NULL;
 
 	SSL_set_verify(tunnel->ssl_handle, SSL_VERIFY_PEER, NULL);
 
 	X509 *cert = SSL_get_peer_certificate(tunnel->ssl_handle);
+
 	if (cert == NULL) {
 		log_error("Unable to get gateway certificate.\n");
 		return 1;
@@ -669,7 +723,7 @@ static int ssl_verify_cert(struct tunnel *tunnel)
 #ifdef HAVE_X509_CHECK_HOST
 	// Use OpenSSL native host validation if v >= 1.0.2.
 	// compare against gateway_host and correctly check return value
-	// to fix piror Incorrect use of X509_check_host
+	// to fix prior incorrect use of X509_check_host
 	if (X509_check_host(cert, tunnel->config->gateway_host,
 	                    0, 0, NULL) == 1)
 		cert_valid = 1;
@@ -710,22 +764,27 @@ static int ssl_verify_cert(struct tunnel *tunnel)
 		goto free_cert;
 	}
 
-	subject = X509_NAME_oneline(subj, NULL, 0);
-	issuer = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
-
-	log_error("Gateway certificate validation failed, and the certificate digest in not in the local whitelist. If you trust it, rerun with:\n");
+	log_error("Gateway certificate validation failed, and the certificate digest is not in the local whitelist. If you trust it, rerun with:\n");
 	log_error("    --trusted-cert %s\n", digest_str);
 	log_error("or add this line to your config file:\n");
 	log_error("    trusted-cert = %s\n", digest_str);
 	log_error("Gateway certificate:\n");
 	log_error("    subject:\n");
-	for (line = strtok(subject, "/"); line != NULL;
-	     line = strtok(NULL, "/"))
-		log_error("        %s\n", line);
+	subject = X509_NAME_oneline(subj, NULL, 0);
+	if (subject) {
+		for (line = strtok_r(subject, "/", &saveptr); line != NULL;
+		     line = strtok_r(NULL, "/", &saveptr))
+			log_error("        %s\n", line);
+		free(subject);
+	}
 	log_error("    issuer:\n");
-	for (line = strtok(issuer, "/"); line != NULL;
-	     line = strtok(NULL, "/"))
-		log_error("        %s\n", line);
+	issuer = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
+	if (issuer) {
+		for (line = strtok_r(issuer, "/", &saveptr); line != NULL;
+		     line = strtok_r(NULL, "/", &saveptr))
+			log_error("        %s\n", line);
+		free(issuer);
+	}
 	log_error("    sha256 digest:\n");
 	log_error("        %s\n", digest_str);
 
@@ -745,7 +804,8 @@ static void ssl_disconnect(struct tunnel *tunnel)
 	SSL_shutdown(tunnel->ssl_handle);
 	SSL_free(tunnel->ssl_handle);
 	SSL_CTX_free(tunnel->ssl_context);
-	close(tunnel->ssl_socket);
+	if (close(tunnel->ssl_socket))
+		log_warn("Could not close ssl socket (%s).\n", strerror(errno));
 
 	tunnel->ssl_handle = NULL;
 	tunnel->ssl_context = NULL;
@@ -762,8 +822,8 @@ int ssl_connect(struct tunnel *tunnel)
 	if (tunnel->ssl_socket == -1)
 		return 1;
 
-	// registration is deprecated from OpenSSL 1.1.0 onward
-#if OPENSSL_API_COMPAT < 0x10100000L
+	// https://wiki.openssl.org/index.php/Library_Initialization
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
 	// Register the error strings for libcrypto & libssl
 	SSL_load_error_strings();
 	// Register the available ciphers and digests
@@ -777,7 +837,7 @@ int ssl_connect(struct tunnel *tunnel)
 		return 1;
 	}
 
-	// Load the OS default CA files
+	/* Load the OS default CA files */
 	if (!SSL_CTX_set_default_verify_paths(tunnel->ssl_context))
 		log_error("Could not load OS OpenSSL files.\n");
 
@@ -791,10 +851,77 @@ int ssl_connect(struct tunnel *tunnel)
 		}
 	}
 
+	/* Disable vulnerable TLS protocols and ciphers by default*/
+	if (!tunnel->config->cipher_list) {
+		if (!tunnel->config->insecure_ssl) {
+			const char *cipher_list;
+
+			if (tunnel->config->seclevel_1)
+				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4@SECLEVEL=1";
+			else
+				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
+			tunnel->config->cipher_list = strdup(cipher_list);
+		} else if (tunnel->config->seclevel_1) {
+			const char *cipher_list = "DEFAULT@SECLEVEL=1";
+
+			tunnel->config->cipher_list = strdup(cipher_list);
+		}
+	}
+
+	if (tunnel->config->cipher_list) {
+		log_debug("Setting cipher list to: %s\n", tunnel->config->cipher_list);
+		if (!SSL_CTX_set_cipher_list(tunnel->ssl_context,
+		                             tunnel->config->cipher_list)) {
+			log_error("SSL_CTX_set_cipher_list failed: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+	}
+
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	if (tunnel->config->min_tls > 0) {
+		log_debug("Setting minimum protocol version to: 0x%x.\n",
+		          tunnel->config->min_tls);
+		if (!SSL_CTX_set_min_proto_version(tunnel->ssl_context,
+		                                   tunnel->config->min_tls)) {
+			log_error("Cannot set minimum protocol version (%s).\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+	}
+#else
+	if (!tunnel->config->insecure_ssl || tunnel->config->min_tls > 0) {
+		long sslctxopt = 0;
+		long checkopt;
+
+		if (!tunnel->config->insecure_ssl)
+			sslctxopt |= SSL_OP_NO_SSLv3 | SSL_OP_NO_COMPRESSION;
+#ifdef TLS1_VERSION
+		if (tunnel->config->min_tls > TLS1_VERSION)
+			sslctxopt |= SSL_OP_NO_TLSv1;
+#endif
+#ifdef TLS1_1_VERSION
+		if (tunnel->config->min_tls > TLS1_1_VERSION)
+			sslctxopt |= SSL_OP_NO_TLSv1_1;
+#endif
+#ifdef TLS1_2_VERSION
+		if (tunnel->config->min_tls > TLS1_2_VERSION)
+			sslctxopt |= SSL_OP_NO_TLSv1_2;
+#endif
+		checkopt = SSL_CTX_set_options(tunnel->ssl_context, sslctxopt);
+		if ((checkopt & sslctxopt) != sslctxopt) {
+			log_error("SSL_CTX_set_options didn't set opt: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+	}
+#endif
+
 	/* Use engine for PIV if user-cert config starts with pkcs11 URI: */
 	if (tunnel->config->use_engine > 0) {
 
 		ENGINE *e;
+
 		ENGINE_load_builtin_engines();
 		e = ENGINE_by_id("pkcs11");
 		if (!e) {
@@ -815,6 +942,7 @@ int ssl_connect(struct tunnel *tunnel)
 		ENGINE_free(e);
 
 		struct token parms;
+
 		parms.uri = tunnel->config->user_cert;
 		parms.cert = NULL;
 
@@ -881,18 +1009,6 @@ int ssl_connect(struct tunnel *tunnel)
 		}
 	}
 
-	if (!tunnel->config->insecure_ssl) {
-		long sslctxopt = SSL_OP_NO_SSLv3 | SSL_OP_NO_COMPRESSION;
-		long checkopt;
-
-		checkopt = SSL_CTX_set_options(tunnel->ssl_context, sslctxopt);
-		if ((checkopt & sslctxopt) != sslctxopt) {
-			log_error("SSL_CTX_set_options didn't set opt: %s\n",
-			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
-		}
-	}
-
 	tunnel->ssl_handle = SSL_new(tunnel->ssl_context);
 	if (tunnel->ssl_handle == NULL) {
 		log_error("SSL_new: %s\n",
@@ -900,49 +1016,6 @@ int ssl_connect(struct tunnel *tunnel)
 		return 1;
 	}
 
-	if (!tunnel->config->insecure_ssl) {
-		if (!tunnel->config->cipher_list) {
-			const char *cipher_list;
-			if (tunnel->config->seclevel_1)
-				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4@SECLEVEL=1";
-			else
-				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
-			tunnel->config->cipher_list = strdup(cipher_list);
-		}
-	} else {
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-		if (tunnel->config->min_tls <= 0)
-			tunnel->config->min_tls = TLS1_VERSION;
-#endif
-		if (!tunnel->config->cipher_list && tunnel->config->seclevel_1) {
-			const char *cipher_list = "DEFAULT@SECLEVEL=1";
-			tunnel->config->cipher_list = strdup(cipher_list);
-		}
-	}
-
-	if (tunnel->config->cipher_list) {
-		log_debug("Setting cipher list to: %s\n", tunnel->config->cipher_list);
-		if (!SSL_set_cipher_list(tunnel->ssl_handle,
-		                         tunnel->config->cipher_list)) {
-			log_error("SSL_set_cipher_list failed: %s\n",
-			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
-		}
-	}
-
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-	if (tunnel->config->min_tls > 0) {
-		log_debug("Setting min proto version to: 0x%x\n",
-		          tunnel->config->min_tls);
-		if (!SSL_set_min_proto_version(tunnel->ssl_handle,
-		                               tunnel->config->min_tls)) {
-			log_error("SSL_set_min_proto_version failed: %s\n",
-			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
-		}
-	}
-#endif
-
 	if (!SSL_set_fd(tunnel->ssl_handle, tunnel->ssl_socket)) {
 		log_error("SSL_set_fd: %s\n",
 		          ERR_error_string(ERR_peek_last_error(), NULL));
