@@ -19,17 +19,18 @@
 #include "userinput.h"
 #include "log.h"
 
+#include <unistd.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <termios.h>
 
 #include <ctype.h>
 #include <errno.h>
 #include <stdarg.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <termios.h>
-#include <unistd.h>
 
 static char *uri_escape(const char *string)
 {
@@ -41,17 +42,23 @@ static char *uri_escape(const char *string)
 	for (i = 0; string[i]; i++) {
 		if (allocated_len + 4 >= real_len) {
 			allocated_len += 16;
-			escaped = realloc(escaped, allocated_len);
+			char *tmp = realloc(escaped, allocated_len);
+
 			// bail out if realloc fails
-			if (escaped == NULL)
-				return NULL;
+			if (tmp == NULL) {
+				free(escaped);
+				escaped = NULL;
+				break;
+			}
+			escaped = tmp;
 		}
 		if (isalnum(string[i]))
 			escaped[real_len++] = string[i];
 		else
 			real_len += sprintf(&escaped[real_len], "%%%02X", string[i]);
 	}
-	escaped[real_len] = '\0';
+	if (escaped)
+		escaped[real_len] = '\0';
 
 	return escaped;
 }
@@ -63,6 +70,10 @@ static char *uri_unescape(const char *string)
 	int real_len = 0;
 	int i = 0;
 
+	// bail out if malloc fails
+	if (unescaped == NULL)
+		return NULL;
+
 	while (string[i]) {
 		if (string[i] == '%' && isxdigit(string[i + 1])
 		    && isxdigit(string[i + 2])) {
@@ -310,6 +321,7 @@ void read_password(const char *pinentry, const char *hint,
 
 	for (i = 0; i < len; i++) {
 		int c = getchar();
+
 		if (c == '\n' || c == EOF)
 			break;
 		pass[i] = (char) c;
